
/** TP 5ETI - CPE Lyon - 2013 */




#include <GL/glew.h>
#include <QString>
#include <QFont>
#include <cmath>
#include <assert.h>

#include "glwidget.hpp"
#include "scene.hpp"
#include "../libmesh/opengl_drawer.hpp"
#include "../libspring/spring_drawer.hpp"
#include "../libperlin/perlin.hpp"
#include "debug_mesh.hpp"



using namespace cpe;

scene::scene()
{
    //Spring parameters
    stiffness=20.0; //stiffness of the spring
    damping=0.1;    //damping=reduction of speed at each time step
    time_step=3e-2; //integration time step

    gravity=vec3(0,0,-9.8);

    //Integration timing parameters
    timer.restart();
    t_current=0.0;
    t_previous=0.0;
    t_diff=0.0;
    framerate=50;

    is_pause=false;

    //Pointers init
    p_widget=NULL;
    p_debug_mesh=NULL;

}

void scene::load_model()
{
    assert(p_widget!=NULL);
    assert(p_debug_mesh!=NULL);


    //*****************************************//
    // Setup Spring                            //
    //*****************************************//

    position_0=vec3(0,0,0);    //first particle position
    position_1=vec3(0,0,-0.5); //second particle position

    speed_0=vec3(0,0,0); //first particle speed
    speed_1=vec3(0,0,0); //second particle speed

    //rest length of the spring
    rest_length=(position_1-position_0).norm();

    //initial displacement
    position_1+=vec3(0.5,0.2,0.5);


    //*****************************************//
    // /Setup Spring                            //
    //*****************************************//



    generate_meshes();

}



void scene::draw_scene()
{
    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);


    //compute spring force at position 1 = spring_force
    cpe::vec3 force=compute_force(stiffness,rest_length,position_0,position_1);
    force += gravity; //add gravity

    //treatment of drawing
    if(is_pause==false)
    {

        t_current=timer.elapsed()/1000.0;
        t_diff += t_current-t_previous;
        t_previous = t_current;

        //integrate at max frequency=framerate
        if(t_diff>= 1.0 / static_cast<float>(framerate) )
        {
            //perform numerical integration
            integration_step(&position_1,&speed_1,force,damping,time_step);

            t_diff=0.0;
        }

    }


    glEnable(GL_LIGHTING);

    //Draw meshes here
    glColor3f(1,1,1);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,texture_id_ground);
    opengl_drawer::draw(ground);
    glDisable(GL_TEXTURE_2D);



    //Draw spring between position_0 & position_1
    glColor3f(1.0,0.0,0.0);
    spring_draw.draw_spring(position_0,position_1);


    //Draw spring force at position_1
    glColor3f(0.1,0.4,0.2);
    glLineWidth(2.0);
    glDisable(GL_LIGHTING);
    spring_draw.draw_spring_force(position_1,force,1.0/10.0);
    glLineWidth(1.0);


    //drawing axes (x,y,z)
    draw_axes();

}

//Compute spring force acting on particle p1
cpe::vec3 scene::compute_force(float K,
                             float L0,
                             const cpe::vec3& p0,
                             const cpe::vec3& p1)
{
    const cpe::vec3 u=p1-p0;
    float L=u.norm();

    cpe::vec3 force=K * (L0-L) * u/L;

    return force;
}

//Compute numerical integration step for one particle
void scene::integration_step(cpe::vec3 *position_pointer,
                             cpe::vec3 *speed_pointer,
                             const cpe::vec3& f,
                             float mu,
                             float dt)
{
    cpe::vec3& p=*position_pointer;
    cpe::vec3& v=*speed_pointer;

    v  = (1.0-mu*dt)*v + dt*f;
    p  =             p + dt*v;
}








void scene::draw_axes()
{
    //affiche les axes
    glDisable(GL_LIGHTING);
    glBegin(GL_LINES);
    glColor3d(1,0,0);
    glVertex3d(0,0,0);
    glVertex3d(1.25,0,0);

    glColor3d(0,1,0);
    glVertex3d(0,0,0);
    glVertex3d(0,1.25,0);

    glColor3d(0,0,1);
    glVertex3d(0,0,0);
    glVertex3d(0,0,1.25);
    glEnd();
}




void scene::generate_meshes()
{
    //*****************************************//
    // Generate mesh                           //
    //*****************************************//
    sphere.load_file("data/sphere.off");
    sphere.scale(0.03,0.03,0.03);
    sphere.compute_normal();

    spring_draw=spring_drawer(sphere);


    ground.build_cube();
    ground.scale(2.0,2.0,0.02);
    ground+=vec3(0,0,-2.0);
    ground.compute_normal();




    const std::string texture_path="data/wood_texture.png";
    texture_id_ground=0;
    texture_id_ground=p_widget->bindTexture(QString(texture_path.c_str()),GL_TEXTURE_2D);
    if(texture_id_ground==0)
        std::cout<<"Erreur chargement texture "<<texture_path<<std::endl;
    //*****************************************//
    // \Generate mesh                          //
    //*****************************************//
}

void scene::set_glwidget(glwidget *p_widget_)
{
    p_widget=p_widget_;
}
void scene::set_debug_mesh(cpe::debug_mesh *p_debug_mesh_)
{
    p_debug_mesh=p_debug_mesh_;
}

