/*
**    TP CPE Lyon
**    Copyright (C) 2013 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef VEC4_HPP
#define VEC4_HPP

#include <string>
#include <iostream>



namespace cpe
{
class vec3;
class vec2;

/** \brief Vectors/Points 4D */
class vec4
{
public:

    // ********************************************* //
    // ********************************************* //
    //  CONSTRUCTORS
    // ********************************************* //
    // ********************************************* //

    /** \brief empty constructor */
    vec4();
    /** \brief direct constructor */
    vec4(float x,float y,float z,float w);

    // ********************************************* //
    // ********************************************* //
    //  Get/set
    // ********************************************* //
    // ********************************************* //

    /** \brief get x coordinate */
    float x() const;
    /** \brief get x coordinate */
    float& x();
    /** \brief get y coordinate */
    float y() const;
    /** \brief get y coordinate */
    float& y();
    /** \brief get z coordinate */
    float z() const;
    /** \brief get z coordinate */
    float& z();
    /** \brief get w coordinate */
    float w() const;
    /** \brief get w coordinate */
    float& w();

    /** \brief Access to the k_th entry (k in [0,3]) */
    float operator[](const size_t& k) const;
    /** \brief Access to the k_th entry (k in [0,3]) */
    float& operator[](const size_t& k);
    /** \brief Access to the k_th entry (k in [0,3]) */
    float operator()(const size_t& k) const;
    /** \brief Access to the k_th entry (k in [0,3]) */
    float& operator()(const size_t& k);

    /** \brief set every entry to 0 */
    void set_zero();

    /** \brief fast pointer access
        \warning unprotected access */
    const float *pointer() const;

    // ********************************************* //
    // ********************************************* //
    //  Convert to lower dimension
    // ********************************************* //
    // ********************************************* //

    /** \brief convert to v3 (x,y,z) */
    vec3 to_vec3() const;
    /** \brief convert to v2 (x,y) */
    vec2 to_vec2() const;



    // ********************************************* //
    // ********************************************* //
    //  Math operation
    // ********************************************* //
    // ********************************************* //

    /** \brief perform dot product between two v4 */
    float dot(const vec4& p) const;

    /** \brief get the norm of the vector */
    float norm() const;
    /** \brief get the square norm of the vector */
    float norm2() const;
    /** \brief normalize the vector to unit length */
    vec4 normalized() const;






    // ********************************************* //
    // ********************************************* //
    //  Operator +-*/
    // ********************************************* //
    // ********************************************* //

    /** \brief + operator */
    vec4 operator+(const vec4& p2) const;
    /** \brief + operator */
    vec4 operator+(float s) const;

    /** \brief - operator */
    vec4 operator-(const vec4& p2) const;
    /** \brief - operator */
    vec4 operator-(float s) const;

    /** \brief multiply by a scalar operator */
    vec4 operator*(float s) const;
    /** \brief divide by a scalar operator */
    vec4 operator/(float s) const;

    /** \brief internal + */
    vec4& operator+=(const vec4& p);
    /** \brief internal + */
    vec4& operator+=(float s);
    /** \brief internal - */
    vec4& operator-=(const vec4& p);
    /** \brief internal - */
    vec4& operator-=(float s);
    /** \brief internal * */
    vec4& operator*=(float s);
    /** \brief internal / */
    vec4& operator/=(float s);

    /** \brief unary negation */
    vec4 operator-() const;

    /** \brief does componentwise mutliplication
            \return MC_v2d(x*p.x(),y*p.y(),z*p.z())
        */
    vec4 product_compontentwise(const vec4& p) const;
    /** \brief does componentwise mutliplication

            perform p=(p.x()*p2.x(),p.y()*p2.x())
        */
    vec4& product_compontentwise_internal(const vec4& p);
    /** \brief internal scaling (similar to componentwise) */
    void scale(float sx,float sy,float sz,float sw);


    // ********************************************* //
    // ********************************************* //
    //  Output
    // ********************************************* //
    // ********************************************* //

    /** \brief export the value as string cout<<v4(2,3,6,7) => 2 3 6 7 */
    std::string to_string() const;



private:


    // ********************************************* //
    // ********************************************* //
    //  Internal parameters
    // ********************************************* //
    // ********************************************* //

    /** \brief x coordinate */
    float internal_x;
    /** \brief y coordinate */
    float internal_y;
    /** \brief z coordinate */
    float internal_z;
    /** \brief w coordinate */
    float internal_w;


    // ********************************************* //
    // ********************************************* //
    //  Helper
    // ********************************************* //
    // ********************************************* //

    /** \brief assert that a size_t belongs to [[0,2]] */
    void assert_size(const size_t& k) const;

};

/** \brief + operator */
vec4 operator+(float s,const vec4& p);
/** \brief - operator */
vec4 operator-(float s,const vec4& p);
/** \brief multiply by a scalar operator */
vec4 operator*(float s,const vec4& p);

// ********************************************* //
// ********************************************* //
//  Output
// ********************************************* //
// ********************************************* //

/** \brief output the vector in ostream as (x,y,z) */
std::ostream& operator<<(std::ostream& stream,const vec4& p);



}



#endif
