/*
**    TP 5ETI CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include "spring_mesh.hpp"

#include "../libmesh/mesh.hpp"

#include <cmath>
#include <sstream>

namespace cpe
{


enum type_spring_stiffness {structural_spring=0,shear_spring=1,bending_spring=2};

spring_mesh::spring_mesh(unsigned int N_init,
                         float K_structural_init,
                         float K_shear_init,
                         float K_bending_init,
                         float damping_init,
                         float dt_init,
                         const cpe::vec3& wind_vector_init)
    :size_spring_mesh(N_init),
      stiffness_structural_spring_mesh(K_structural_init),
      stiffness_shear_spring_mesh(K_shear_init),
      stiffness_bending_spring_mesh(K_bending_init),
      damping_spring_mesh(damping_init),
      time_step_spring_mesh(dt_init),
      wind_vector_spring_mesh(wind_vector_init)
{
    init();
}



void spring_mesh::compute_force()
{
    //security check for size
    unsigned int N_vertices=size_spring_mesh*size_spring_mesh;
    if(forces_spring_mesh.size()!=N_vertices)
        forces_spring_mesh.resize(N_vertices);

    if(forces_spring_mesh.size()!=N_vertices)
        throw cpe::exception_cpe("Bad size of forces",MACRO_EXCEPTION_PARAMETER);


    //set force to gravity
    const cpe::vec3 g(0,0,-9.8);
    for(unsigned int k=0;k<N_vertices;++k)
        forces_spring_mesh[k]=g/static_cast<float>(N_vertices);




    /********************************************/
    /********************************************/
    /* TO DO: Force computation                 */
    /* ...                                      */
    /* ...                                      */
    /* ...                                      */
    /* ...                                      */
    /* ...                                      */
    /********************************************/
    /********************************************/

}


void spring_mesh::numerical_integration()
{
    //security check
    unsigned int N_vertices=size_spring_mesh*size_spring_mesh;
    if(speed_spring_mesh.size()!=N_vertices)
        speed_spring_mesh.resize(N_vertices);

    if(speed_spring_mesh.size()!=number_of_vertices())
        throw cpe::exception_cpe("Bad size of speed",MACRO_EXCEPTION_PARAMETER);





    /********************************************/
    /********************************************/
    /* TO DO: numerical_integration             */
    /* ...                                      */
    /* ...                                      */
    /* ...                                      */
    /* ...                                      */
    /* ...                                      */
    /* ...                                      */
    /********************************************/
    /********************************************/






    //security check (throw exception if divergence is detected)
    const float LIMIT=100.0;
    for(unsigned int k=0;k<N_vertices;++k)
    {
        const vec3& p=mesh::vertex(k);

        if(p.norm2()>LIMIT)
        {
            throw cpe::exception_cpe("Divergence of the system ",MACRO_EXCEPTION_PARAMETER);
            init();
        }
    }


}



void spring_mesh::init()
{
    if(size_spring_mesh==0)
        return ;

    unsigned int N_vertices=size_spring_mesh*size_spring_mesh;
    if(mesh::number_of_vertices()!=N_vertices)
        mesh::get_vertices().resize(N_vertices);
    if(mesh::get_texture().size()!=N_vertices)
        mesh::get_texture().resize(N_vertices);

    if(forces_spring_mesh.size()!=N_vertices)
        forces_spring_mesh.resize(N_vertices);
    if(speed_spring_mesh.size()!=N_vertices)
        speed_spring_mesh.resize(N_vertices);


    if(size_spring_mesh<=0)
        throw cpe::exception_cpe("Bad size of mesh",MACRO_EXCEPTION_PARAMETER);
    if(mesh::number_of_vertices()!=N_vertices)
        throw cpe::exception_cpe("Error size of mesh",MACRO_EXCEPTION_PARAMETER);
    if(mesh::get_texture().size()!=N_vertices)
        throw cpe::exception_cpe("Bad size of texture",MACRO_EXCEPTION_PARAMETER);
    if(speed_spring_mesh.size()!=N_vertices)
        throw cpe::exception_cpe("Bad size of speed",MACRO_EXCEPTION_PARAMETER);
    if(forces_spring_mesh.size()!=N_vertices)
        throw cpe::exception_cpe("Bad size of forces",MACRO_EXCEPTION_PARAMETER);




    for(unsigned int kx=0;kx<size_spring_mesh;++kx)
    {
        float ux=static_cast<float>(kx)/size_spring_mesh;

        for(unsigned int ky=0;ky<size_spring_mesh;++ky)
        {
            float uy=static_cast<float>(ky)/size_spring_mesh;

            mesh::vertex(coord_to_index(kx,ky)) = cpe::vec3(ux,uy,0.0);
            mesh::texture(coord_to_index(kx,ky)) = cpe::vec2(ux,uy);

        }
    }

    unsigned int N_triangle=2*(size_spring_mesh-1)*(size_spring_mesh-1);
    if(get_connectivity().size()!=3*N_triangle)
        get_connectivity().resize(3*N_triangle);

    if(number_of_triangle()!=N_triangle)
        throw cpe::exception_cpe("Bad number of triangles",MACRO_EXCEPTION_PARAMETER);


    unsigned int counter=0;
    for(unsigned int kx=0;kx<size_spring_mesh-1;++kx)
    {
        for(unsigned int ky=0;ky<size_spring_mesh-1;++ky)
        {
            connectivity(3*counter+0)=coord_to_index(kx+0,ky+0);
            connectivity(3*counter+1)=coord_to_index(kx+1,ky+0);
            connectivity(3*counter+2)=coord_to_index(kx+1,ky+1);
            counter++;

            connectivity(3*counter+0)=coord_to_index(kx+0,ky+0);
            connectivity(3*counter+1)=coord_to_index(kx+1,ky+1);
            connectivity(3*counter+2)=coord_to_index(kx+0,ky+1);
            counter++;

        }
    }


    if(counter!=N_triangle)
        throw cpe::exception_cpe("Bad counter",MACRO_EXCEPTION_PARAMETER);


    for(unsigned int k=0;k<N_vertices;++k)
    {
        speed_spring_mesh[k]=vec3(0,0,0);
        forces_spring_mesh[k]=vec3(0,0,0);
    }


    mesh::compute_normal();



}

unsigned int spring_mesh::coord_to_index(unsigned int kx,unsigned int ky) const
{
    if(kx>size_spring_mesh || ky>size_spring_mesh)
    {
        std::stringstream msg;
        msg<<"Index exceed size of the spring mesh ("<<kx<<","<<ky<<") > ("<<size_spring_mesh<<","<<size_spring_mesh<<")"<<std::endl;
        throw cpe::exception_cpe(msg.str(),MACRO_EXCEPTION_PARAMETER);
    }
    return ky+size_spring_mesh*kx;
}



unsigned int spring_mesh::size() const{return size_spring_mesh;}



vec3& spring_mesh::vertex(unsigned int kx,unsigned int ky)
{
    unsigned int index=coord_to_index(kx,ky);

    if(index>=number_of_vertices())
        throw cpe::exception_cpe("Index of mesh too large",MACRO_EXCEPTION_PARAMETER);

    return mesh::vertex(index);
}

const vec3& spring_mesh::vertex(unsigned int kx,unsigned int ky) const
{
    unsigned int index=coord_to_index(kx,ky);

    if(index>=number_of_vertices())
        throw cpe::exception_cpe("Index of mesh too large",MACRO_EXCEPTION_PARAMETER);

    return mesh::vertex(index);
}

cpe::vec3& spring_mesh::force(unsigned int kx,unsigned int ky)
{
    unsigned int index=coord_to_index(kx,ky);

    if(index>=forces_spring_mesh.size())
        throw cpe::exception_cpe("Index of force too large",MACRO_EXCEPTION_PARAMETER);

    return forces_spring_mesh[index];
}

const cpe::vec3& spring_mesh::force(unsigned int kx,unsigned int ky) const
{
    unsigned int index=coord_to_index(kx,ky);

    if(index>=forces_spring_mesh.size())
        throw cpe::exception_cpe("Index of force too large",MACRO_EXCEPTION_PARAMETER);

    return forces_spring_mesh[index];
}

cpe::vec3& spring_mesh::speed(unsigned int kx,unsigned int ky)
{
    unsigned int index=coord_to_index(kx,ky);

    if(index>=speed_spring_mesh.size())
        throw cpe::exception_cpe("Index of speed too large",MACRO_EXCEPTION_PARAMETER);

    return speed_spring_mesh[index];
}

const cpe::vec3& spring_mesh::speed(unsigned int kx,unsigned int ky) const
{
    unsigned int index=coord_to_index(kx,ky);
    if(index>=speed_spring_mesh.size())
        throw cpe::exception_cpe("Index of speed too large",MACRO_EXCEPTION_PARAMETER);

    return speed_spring_mesh[index];
}


float& spring_mesh::K_structural(){return stiffness_structural_spring_mesh;}
const float& spring_mesh::K_structural() const{return stiffness_structural_spring_mesh;}
float& spring_mesh::K_shear(){return stiffness_shear_spring_mesh;}
const float& spring_mesh::K_shear() const{return stiffness_shear_spring_mesh;}
float& spring_mesh::K_bending(){return stiffness_bending_spring_mesh;}
const float& spring_mesh::K_bending() const{return stiffness_bending_spring_mesh;}
float& spring_mesh::damping(){return damping_spring_mesh;}
const float& spring_mesh::damping() const{return damping_spring_mesh;}

void spring_mesh::update_normal()
{
    const unsigned int N_vertices=size_spring_mesh*size_spring_mesh;
    if(mesh::get_normal().size()!=N_vertices)
        mesh::get_normal().resize(N_vertices);

    if(mesh::get_normal().size()!=number_of_vertices())
        throw cpe::exception_cpe("Bad number of normals",MACRO_EXCEPTION_PARAMETER);


    for(unsigned int kx=1;kx<size_spring_mesh-1;++kx)
    {
        for(unsigned int ky=1;ky<size_spring_mesh-1;++ky)
        {
            const vec3& p0=vertex(kx,ky);

            const vec3& pxp=vertex(kx+1,ky);
            const vec3& pxm=vertex(kx-1,ky);
            const vec3& pyp=vertex(kx,ky+1);
            const vec3& pym=vertex(kx,ky-1);

            const vec3 dxp=(pxp-p0).normalized();
            const vec3 dxm=(p0-pxm).normalized();

            const vec3 dyp=(pyp-p0).normalized();
            const vec3 dym=(p0-pym).normalized();

            const vec3 dx=(dxp+dxm).normalized();
            const vec3 dy=(dyp+dym).normalized();

            const vec3 dn=(dx.cross(dy)).normalized();

            normal(coord_to_index(kx,ky))=dn;
        }
    }

    for(unsigned int k=1;k<size_spring_mesh-1;++k)
    {
        normal(coord_to_index(k,0))=normal(coord_to_index(k,1));
        normal(coord_to_index(k,size_spring_mesh-1))=normal(coord_to_index(k,size_spring_mesh-2));

        normal(coord_to_index(0,k))=normal(coord_to_index(1,k));
        normal(coord_to_index(size_spring_mesh-1,k))=normal(coord_to_index(size_spring_mesh-2,k));
    }

    normal(coord_to_index(0,0))=normal(coord_to_index(1,1));
    normal(coord_to_index(0,size_spring_mesh-1))=normal(coord_to_index(1,size_spring_mesh-2));

    normal(coord_to_index(size_spring_mesh-1,0))=normal(coord_to_index(size_spring_mesh-2,1));
    normal(coord_to_index(size_spring_mesh-1,size_spring_mesh-1))=normal(coord_to_index(size_spring_mesh-2,size_spring_mesh-2));
}

void spring_mesh::init(unsigned int s)
{
    size_spring_mesh=s;
    init();

}
float& spring_mesh::time_step() {return time_step_spring_mesh;}
const float& spring_mesh::time_step() const{return time_step_spring_mesh;}

cpe::vec3& spring_mesh::wind_vector()
{
    return wind_vector_spring_mesh;
}

const cpe::vec3& spring_mesh::wind_vector() const
{
    return wind_vector_spring_mesh;
}


}
