
/*
**    TP 5ETI CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef SPRING_MESH_HPP
#define SPRING_MESH_HPP

#include <iostream>
#include <vector>

#include "../libmesh/mesh.hpp"
#include "../lib3d/vec3.hpp"


namespace cpe
{




    /** \brief square spring mesh */
    class spring_mesh : public cpe::mesh
    {
    public:

        // ********************************************* //
        // ********************************************* //
        //  CONSTRUCTORS
        // ********************************************* //
        // ********************************************* //

        /** \brief constructor */
        spring_mesh(unsigned int N=0,
                    float K_structural=1.0,
                    float K_shear=1.0,
                    float K_bending=1.0,
                    float damping=0.1,
                    float dt=1e-2,
                    const cpe::vec3& wind_vector=cpe::vec3(0,0,0));


        // ********************************************* //
        // ********************************************* //
        //  Dynamic
        // ********************************************* //
        // ********************************************* //

        /** \brief compute force */
        void compute_force();

        /** \brief apply force with stepsize dt */
        void numerical_integration();


        // ********************************************* //
        // ********************************************* //
        //  update visualisation
        // ********************************************* //
        // ********************************************* //

        /** \brief Compute normal for the mesh */
        void update_normal();

        // ********************************************* //
        // ********************************************* //
        //  Init
        // ********************************************* //
        // ********************************************* //

        /** \brief init values of positions */
        void init();
        /** \brief init values of positions with given new size */
        void init(unsigned int s);


        // ********************************************* //
        // ********************************************* //
        //  Size and index
        // ********************************************* //
        // ********************************************* //

        /** \brief size N (number of vertex per row: total vertices=NxN) */
        unsigned int size() const;


        /** \brief convert 2D coordinate (kx,ky) into 1D index
            index = ky + size_y*kx
        */
        unsigned int coord_to_index(unsigned int kx,unsigned int ky) const;


        // ********************************************* //
        // ********************************************* //
        //  Element Access
        // ********************************************* //
        // ********************************************* //



        /** \brief access to vertex (as parametric structure) */
        cpe::vec3& vertex(unsigned int kx,unsigned int ky);
        /** \brief access to vertex (as parametric structure) */
        const cpe::vec3& vertex(unsigned int kx,unsigned int ky) const;

        /** \brief access to force (as parametric structure) */
        cpe::vec3& force(unsigned int kx,unsigned int ky);
        /** \brief access to force (as parametric structure) */
        const cpe::vec3& force(unsigned int kx,unsigned int ky) const;

        /** \brief access to speed (as parametric structure) */
        cpe::vec3& speed(unsigned int kx,unsigned int ky);
        /** \brief access to speed (as parametric structure) */
        const cpe::vec3& speed(unsigned int kx,unsigned int ky) const;


        // ********************************************* //
        // ********************************************* //
        //  attribute
        // ********************************************* //
        // ********************************************* //



        /** \brief get structural stiffness */
        float& K_structural();
        /** \brief get structural stiffness */
        const float& K_structural() const;

        /** \brief get shear stiffness */
        float& K_shear();
        /** \brief get shear stiffness */
        const float& K_shear() const;

        /** \brief get bending stiffness */
        float& K_bending();
        /** \brief get bending stiffness */
        const float& K_bending() const;

        /** \brief get damping coeff */
        float& damping();
        /** \brief get damping coeff */
        const float& damping() const;


        /** \brief get time step */
        float& time_step();
        /** \brief get time step */
        const float& time_step() const;

        /** \brief get wind vector */
        cpe::vec3& wind_vector();
        /** \brief get wind vector */
        const cpe::vec3& wind_vector() const;




    private:

        /** \brief number of vertices = NxN */
        unsigned int size_spring_mesh;





        /** \brief force container */
        std::vector<cpe::vec3> forces_spring_mesh;
        /** \brief speed container */
        std::vector<cpe::vec3> speed_spring_mesh;


        /** \brief structural stiffness */
        float stiffness_structural_spring_mesh;
        /** \brief shear stiffness */
        float stiffness_shear_spring_mesh;
        /** \brief bending stiffness */
        float stiffness_bending_spring_mesh;

        /** \brief damping coefficient */
        float damping_spring_mesh;
        /** \brief time step */
        float time_step_spring_mesh;

        /** \brief wind vector */
        cpe::vec3 wind_vector_spring_mesh;

    };
}

#endif
