
/** TP 5ETI - CPE Lyon - 2013 */

#include <GL/glew.h>

#include "scene.hpp"
#include "glwidget.hpp"

#include "../libmesh/opengl_drawer.hpp"
#include "../libspring/spring_drawer.hpp"
#include "../libperlin/perlin.hpp"
#include "debug_mesh.hpp"


#include <QString>
#include <QFont>
#include <cmath>
#include <cassert>

using namespace cpe;

//setup init parameters
scene::scene()
{

    //Integration timing parameters
    timer.restart();
    t_current=0.0;
    t_previous=0.0;
    t_diff=0.0;
    framerate=50;

    is_pause=false;
    is_grid=false;
    is_surface=true;
    steps_cloth=5;
    delta_grid=5;

    //Pointers init
    p_widget=NULL;
    p_debug_mesh=NULL;

}


//load init scene
void scene::load_model()
{
    assert(p_widget!=NULL);
    assert(p_debug_mesh!=NULL);

    //cloth init
    cloth=spring_mesh(50,20.0,20.0,20.0,0.2,0.1,cpe::vec3(0.5,0,0));
    cloth.compute_normal();




    generate_meshes();

}


//drawing loop
void scene::draw_scene()
{
    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);



    //treatment of drawing
    if(is_pause==false) //do not integrate during pause
    {

        t_current=timer.elapsed()/1000.0;
        t_diff += t_current-t_previous;
        t_previous = t_current;

        //integrate at max frequency=framerate
        if(t_diff>= 1.0 / static_cast<float>(framerate) )
        {
            try
            {

                for(unsigned int k=0;k<steps_cloth;++k) //force computation and numerical integration
                {
                    cloth.compute_force();
                    cloth.numerical_integration();
                    cloth.update_normal();
                }
            }
            catch(cpe::exception_cpe& e) //stop the simulation of something bad happen
            {
                std::cout<<"==============================================="<<std::endl;
                std::cout<<"Catch exception in numerical integration loop"  <<std::endl;
                std::cout<<e.info()<<std::endl;
                std::cout<<"==============================================="<<std::endl;

                set_is_pause(true);
            }


            t_diff=0.0;
        }

    }


    glEnable(GL_LIGHTING);

    //Draw meshes here
    glColor3f(1,1,1);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,texture_id_ground);
    opengl_drawer::draw(ground);
    glDisable(GL_TEXTURE_2D);

    opengl_drawer::draw(ball);



    //drawing axes (x,y,z)
    draw_axes();

    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glColor3f(0,0,0);
    if(is_grid==true)
        draw_grid(cloth);


    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(2,2);

    glEnable(GL_LIGHTING);
    glColor3f(1.0,1.0,1.0);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,texture_id_cloth);
    if(is_surface==true)
        opengl_drawer::draw(cloth);


    p_debug_mesh->draw_debug(cloth);


    glDisable(GL_POLYGON_OFFSET_FILL);


}




void scene::draw_axes()
{
    //affiche les axes
    glDisable(GL_LIGHTING);
    glBegin(GL_LINES);
    glColor3d(1,0,0);
    glVertex3d(0,0,0);
    glVertex3d(1.25,0,0);

    glColor3d(0,1,0);
    glVertex3d(0,0,0);
    glVertex3d(0,1.25,0);

    glColor3d(0,0,1);
    glVertex3d(0,0,0);
    glVertex3d(0,0,1.25);
    glEnd();
}




void scene::generate_meshes()
{
    //*****************************************//
    // Generate mesh                           //
    //*****************************************//
    sphere.load_file("data/sphere.off");
    sphere.scale(0.03,0.03,0.03);
    sphere.compute_normal();


    ground.build_cube();
    float L=0.02;
    ground.scale(2.0,2.0,L);
    ground+=vec3(0,0,-0.9-2*L);
    ground.compute_normal();


    ball.load_file("data/sphere.off");
    ball.scale(0.1,0.1,0.1);
    ball+=cpe::vec3(0.5,0,-0.8);
    ball.compute_normal();

    const std::string texture_path="data/wood_texture.png";
    texture_id_ground=0;
    texture_id_ground=p_widget->bindTexture(QString(texture_path.c_str()),GL_TEXTURE_2D);
    if(texture_id_ground==0)
        std::cout<<"Erreur chargement texture "<<texture_path<<std::endl;

    const std::string texture_path_cloth="data/cloth.png";
    texture_id_cloth=0;
    texture_id_cloth=p_widget->bindTexture(QString(texture_path_cloth.c_str()),GL_TEXTURE_2D);
    if(texture_id_cloth==0)
        std::cout<<"Erreur chargement texture "<<texture_path_cloth<<std::endl;

    //*****************************************//
    // \Generate mesh                          //
    //*****************************************//
}

void scene::set_glwidget(glwidget *p_widget_)
{
    p_widget=p_widget_;
}
void scene::set_debug_mesh(cpe::debug_mesh *p_debug_mesh_)
{
    p_debug_mesh=p_debug_mesh_;
}

void scene::set_is_pause(bool is_enable)
{
    is_pause=is_enable;
}
void scene::set_is_grid(bool is_enable)
{
    is_grid=is_enable;
}
void scene::set_is_surface(bool is_enable)
{
    is_surface=is_enable;
}
void scene::set_stiffness_structural(float K_struct)
{
    cloth.K_structural()=K_struct;
}
void scene::set_stiffness_shear(float K_shear)
{
    cloth.K_shear()=K_shear;
}
void scene::set_stiffness_bending(float K_bending)
{
    cloth.K_bending()=K_bending;
}
void scene::set_size_cloth(int s)
{
    cloth.init(s);
}
void scene::set_time_step_cloth(float dt)
{
    cloth.time_step()=dt;
}
void scene::set_damping_cloth(float mu)
{
    cloth.damping()=mu;
}
void scene::set_steps(int steps)
{
    steps_cloth=steps;
}
void scene::restart()
{
    cloth.init();
}
void scene::set_delta_grid(int d_grid)
{
    delta_grid=d_grid;
}

void scene::set_wind(const cpe::vec3& w)
{
    cloth.wind_vector()=w;
}

void scene::draw_grid(const cpe::spring_mesh& mesh)
{
    glLineWidth(4.0);
    unsigned int N=mesh.size();
    glBegin(GL_LINE_STRIP);
    for(unsigned int k=0;k<N;++k)
    {
        const cpe::vec3& p=mesh.vertex(k,0);
        glVertex3fv(p.pointer());
    }
    glEnd();

    glBegin(GL_LINE_STRIP);
    for(unsigned int k=0;k<N;++k)
    {
        const cpe::vec3& p=mesh.vertex(k,N-1);
        glVertex3fv(p.pointer());
    }
    glEnd();

    glBegin(GL_LINE_STRIP);
    for(unsigned int k=0;k<N;++k)
    {
        const cpe::vec3& p=mesh.vertex(0,k);
        glVertex3fv(p.pointer());
    }
    glEnd();

    glBegin(GL_LINE_STRIP);
    for(unsigned int k=0;k<N;++k)
    {
        const cpe::vec3& p=mesh.vertex(N-1,k);
        glVertex3fv(p.pointer());
    }
    glEnd();


    glLineWidth(2.0);


    for(unsigned int ks=delta_grid;ks<N;ks+=delta_grid)
    {
        glBegin(GL_LINE_STRIP);
        for(unsigned int k=0;k<N;++k)
        {
            const cpe::vec3& p=mesh.vertex(k,ks);
            glVertex3fv(p.pointer());
        }
        glEnd();

        glBegin(GL_LINE_STRIP);
        for(unsigned int k=0;k<N;++k)
        {
            const cpe::vec3& p=mesh.vertex(ks,k);
            glVertex3fv(p.pointer());
        }
        glEnd();
    }

    glLineWidth(1.0);
}
